
import time
import queue
import threading

from ..connection import Connection
from ..exceptions import getExceptionFromError

class GetManyMany:

    class Result:

        def __init__(self, tree, shot, result):
            self._tree = tree
            self._shot = shot
            self._result = result

        @property
        def tree(self):
            return self._tree

        @property
        def shot(self):
            return self._shot

        def get(self, name):
            
            if name not in self._result:
                return None
            
            result = self._result[name]
            if 'value' in result:
                return result['value']
            
            raise getExceptionFromError(result['error'].data())

    class Worker(threading.Thread):

        def __init__(self, gmm):
            super().__init__()

            self._gmm = gmm

        def run(self):

            c = Connection(self._gmm._connection_url, **self._gmm._connection_kwargs)

            while True:
                try:
                    tree, shot = self._gmm._shots.get_nowait()
                except queue.Empty:
                    break

                gm = c.getMany()
                gm.append('_gmm_open', 'TreeOpen($,$)', tree, shot)

                for query in self._gmm._queries:
                    gm.append(query['name'], query['exp'], *query['args'])

                result = gm.execute()
                
                self._gmm._results.put(GetManyMany.Result(tree, shot, result))

            c.disconnect()

    def __init__(self, connection_url: str, num_workers: int = 8, worker_delay: float = 0.0, **connection_kwargs):
        
        self._connection_url = connection_url
        self._connection_kwargs = connection_kwargs
        self._num_workers = num_workers
        self._worker_delay = worker_delay

        self._total_shots = 0
        self._shots = queue.Queue()
        self._results = queue.Queue()
        self._queries = []
        self._workers = []

    def add_shots(self, tree, shots):
        if not isinstance(shots, list):
            shots = [ shots ]
        
        for shot in shots:
            self._shots.put((tree, shot))
            self._total_shots += 1

    def append(self, name, exp, *args):
        self._queries.append({
            'name': name,
            'exp': exp,
            'args': list(args),
        })

    def execute(self):
        
        num_workers = min(self._num_workers, self._total_shots)
        for _ in range(num_workers):
            worker = self.Worker(self)
            worker.start()

            if self._worker_delay > 0:
                time.sleep(self._worker_delay)

            self._workers.append(worker)

            # If the previous threads have already finished, don't make new ones
            if self._shots.qsize() == 0:
                break
        
        for _ in range(self._total_shots):
            yield self._results.get()
        
        for worker in self._workers:
            worker.join()



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import ctypes
import numpy

from .exceptions import *
from .internals.dtypedef import *
from .internals.classdef import *
from .internals.mdsdescrip import *
from .internals.CvtConvertFloat import convert_float, convert_float_array

###
### Numeric
###

class Numeric:
    """
    Numeric type base class

    This contains the useful mathematical operators for all numeric types
    """

    def __new__(cls, *args, **kwargs):
        if cls is Numeric:
            raise Exception(f'Numeric cannot be instantiated directly')
        
        return object.__new__(cls)
    
    @property
    def __array_interface__(self):
        return self._data.__array_interface__
    
    def __int__(self):
        return int(self._data)
    
    def __float__(self):
        return float(self._data)
    
    def __add__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data + other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data + other)
    
    def __sub__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data - other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data - other)
    
    def __mul__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data * other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data * other)
    
    def __pow__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data ** other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data ** other)
    
    def __truediv__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data / other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data / other)
    
    def __floordiv__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data // other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data // other)
    
    def __mod__(self, other):
        if isinstance(other, Signal):
            return Signal(self._data % other.data(), None, *other.dimensions)
        return Descriptor.from_data(self._data % other)
    
    def __lshift__(self, other):
        return self.__class__(self._data << other)
    
    def __rshift__(self, other):
        return self.__class__(self._data >> other)
    
    def __and__(self, other):
        return self.__class__(self._data & other)
    
    def __or__(self, other):
        return self.__class__(self._data | other)
    
    def __xor__(self, other):
        return self.__class__(self._data ^ other)
    
    def __not__(self):
        return self.__class__(~self._data)
    
###
### Descriptor
###

class Descriptor:
    """
    Descriptor base class, which provides several useful methods for working with Descriptors.
    
    If called with no parameters, e.g. :class:`Descriptor()`, this will return the NULL Descriptor.
    
    If called with a python or numpy data type, it will coerce it into the proper MDSplus Descriptor
    subclass. e.g.:

    .. code-block:: python

        Descriptor(42) -> Int64(42)
        Descriptor(numpy.array([1, 2, 3], dtype=numpy.uint8)) -> UInt8Array([1, 2, 3])

    If you want to ensure that a given piece of data is a :class:`Descriptor`, call :meth:`Descriptor.from_data()`
    which will convert it if needed, or return the original :class:`Descriptor` if it already is one. e.g.

    .. code-block:: python

        Descriptor.from_data(42) -> Int64(42)
        Descriptor.from_data(Action(1, 2, 3)) -> Action(1, 2, 3)

    This also provides the :meth:`Descriptor.unpack()` and :meth:`Descriptor.pack()` methods, which are then
    overridden by each subclass as necessary. Also, the :meth:`Descriptor.serialize()` method will call
    :meth:`Descriptor.pack()`, but return the data as a `UInt8Array`. You could then call the
    :meth:`UInt8Array.deserialize()` method on either :class:`Int8Array` or :class:`UInt8Array` types.

    The `repr()` of a :class:`Descriptor` [sub]class should give you the TDI Decompile of that type. e.g.

    .. code-block:: python

        repr(Descriptor()) -> '*'
        repr(Int32(42)) -> '42L'
        repr(Signal([1, 2, 3], None, [4, 5, 6])) -> 'Build_Signal(Quadword([1, 2, 3]), *, Quadword([4, 5, 6]))'

    """

    def __new__(cls, data=None, dsc=None, *args, **kwargs):

        if cls is Descriptor:

            if dsc is not None:
                cls = DTYPE_CLASS_MAP[dsc.class_id][dsc.dtype_id]

            else:
                if isinstance(data, Descriptor):
                    cls = type(data)

                elif isinstance(data, str):
                    cls = String
                elif isinstance(data, bool):
                    cls = UInt8
                elif isinstance(data, int):
                    cls = Int64
                elif isinstance(data, float):
                    cls = Float64

                elif isinstance(data, (bytes, bytearray, memoryview)):
                    cls = UInt8Array
                elif isinstance(data, list):
                    cls = List
                elif isinstance(data, tuple):
                    cls = Tuple
                elif isinstance(data, dict):
                    cls = Dictionary

                elif isinstance(data, numpy.number):
                    
                    for dtype_id, numpy_dtype in NUMPY_DTYPE_MAP.items():
                        if data.dtype == numpy_dtype:
                            cls = DTYPE_CLASS_MAP[CLASS_S][dtype_id]
                            break

                elif isinstance(data, numpy.ndarray):
                    
                    for dtype_id, numpy_dtype in NUMPY_DTYPE_MAP.items():
                        if data.dtype == numpy_dtype:
                            cls = DTYPE_CLASS_MAP[CLASS_A][dtype_id]
                            break

                    if data.dtype.char in ['U', 'S']:
                        cls = StringArray
        
        if cls is Descriptor:
            return object.__new__(cls)
        
        return cls.__new__(cls)

    def __init__(self, data=None, dsc=None):
        self._dsc = dsc
        self._data = data

        if self._dsc is None:
            self._dsc = mdsdsc_t(
                length=0,
                class_id=CLASS_MISSING,
                dtype_id=DTYPE_MISSING,
                offset=0
            )

    @property
    def length(self):
        """
        :class:`mdsdsc_t`.length
        """

        return self._dsc.length
    
    @property
    def dtype_id(self):
        """
        :class:`mdsdsc_t`.dtype_id
        """

        return self._dsc.dtype_id

    @property
    def dtype_str(self):
        """
        dtype_to_string(:class:`mdsdsc_t`.dtype_id)
        """

        return dtype_to_string(self._dsc.dtype_id)
    
    @property
    def class_id(self):
        """
        :class:`mdsdsc_t`.class_id
        """

        return self._dsc.class_id
    
    @property
    def class_str(self):
        """
        class_to_string(:class:`mdsdsc_t`.class_id)
        """

        return class_to_string(self._dsc.class_id)

    @property
    def offset(self):
        """
        :class:`mdsdsc_t`.offset
        """

        return self._dsc.offset

    def data(self):
        """
        Get the data for this descriptor

        :return: Return the data in this descriptor as a python/numpy type.
        :rtype: The python/numpy type of the underlying data, overridden by all subclasses.
        """

        return self._data
    
    def decompile(self):
        """
        Return the decompiled TDI representation of the data.
        """

        return repr(self)

    def __repr__(self):
        if self._data is None:
            return '*'
        return f'{self.__class__.__name__}({self._data})'
    
    def __eq__(self, other):
        if isinstance(other, Descriptor):
            other = other.data()
        return self.data() == other
    
    def __hash__(self):
        return hash(self.data())
    
    def serialize(self):
        return UInt8Array(self.pack())
    
    @staticmethod
    def from_data(data):
        """
        Ensure that the given data is a :class:`Descriptor` type. If the data is already
        a :class:`Descriptor`, then just return it. Otherwise, call :class:`Descriptor(data)`

        :param data: The data to possibly convert into a :class:`Descriptor`
        :type data: python/numpy type or :class:`Descriptor`
        :return: The data as a :class:`Descriptor`
        :rtype: Descriptor
        """

        if isinstance(data, Descriptor):
            return data
        
        return Descriptor(data)
    
    def pack(self):
        """
        Pack the :class:`Descriptor` into a serialized array of bytes, starting with the `mdsdsc_t` header.
        This is overridden by all subclasses.

        :return: The :class:`Descriptor` serialized into bytes.
        :rtype: `bytearray()`
        """
        return bytearray()
    
    @staticmethod
    def unpack(buffer, conn=None):
        """
        Unpack the given buffer and construct the corresponding :class:`Descriptor` subclass. This requires a
        buffer with a `mdsdsc_t` header.

        :param buffer: The buffer to unpack.
        :type buffer: bytes, bytearray or any type that implements the buffer protocol.
        :param Connection conn: The connection, used to gather missing metadata, such as the FULLPATH
            of a given NID, defaults to None
        :return: An instance of a :class:`Descriptor` subclass containing the data.
        :rtype: A subclass of :class:`Descriptor`
        :raises MdsException: if there are problems unpacking the data.
        """

        buffer = bytearray(buffer)
        
        # TODO: Improve?
        dtype_id = buffer[2]
        class_id = buffer[3]

        if class_id not in DTYPE_CLASS_MAP:
            raise MdsException('Invalid class:', class_to_string(class_id))
        
        if dtype_id not in DTYPE_CLASS_MAP[class_id]:
            raise MdsException('Invalid dtype for class:', dtype_to_string(dtype_id), class_to_string(class_id))

        dtype_class = DTYPE_CLASS_MAP[class_id][dtype_id]

        if issubclass(dtype_class, DescriptorS):

            dsc = mdsdsc_s_t.from_buffer(buffer)

            if dsc.length == 0:
                dsc.length = get_dtype_size(dsc.dtype_id)

            if dsc.offset == 0:
                dsc.offset = ctypes.sizeof(dsc)

            data = None
            data_buffer = buffer[ dsc.offset : dsc.offset + dsc.length ]
            
            if dsc.dtype_id in NUMPY_DTYPE_MAP:
                data = numpy.frombuffer(data_buffer, dtype=NUMPY_DTYPE_MAP[dtype_id], count=1)[0]
            
            elif dsc.dtype_id == DTYPE_T:
                data = data_buffer.decode('ascii')
            
            elif dsc.dtype_id in [ DTYPE_F, DTYPE_D, DTYPE_G ]:
                data = convert_float(dsc.dtype_id, data_buffer)

            if dtype_class is TreeNID:
                return TreeNID(data, conn=conn)
                
            return dtype_class(data)

        elif issubclass(dtype_class, DescriptorA):

            dsc = mdsdsc_a_t.from_buffer(buffer)

            if dsc.length == 0:
                dsc.length = get_dtype_size(dsc.dtype_id)

            count = dsc.arsize // dsc.length
            shape = (count,)
            order = 'C' # C-style Row-Major

            if dsc.scale > 0:
                raise MdsException('Array scale unimplemented')
            
            if dsc.digits > 0:
                raise MdsException('Array digits unimplemented')
            
            if dsc.aflags.binscale:
                raise MdsException('Array binscale unimplemented')
            
            # if dsc.aflags.redim:
            #     raise MdsException('Array redim unimplemented')

            if dsc.aflags.coeff:

                a0_coeff_buffer = buffer[ ctypes.sizeof(dsc) : ]
                a0_coeff = numpy.frombuffer(a0_coeff_buffer, dtype='uint32', count=(1 + dsc.dimct))
                
                a0 = a0_coeff[0]
                coeff = a0_coeff[1 : ]

                dsc.offset = a0 # "address of element whos index is all zeros"
                shape = coeff[ : : -1]

                if dsc.aflags.bounds:
                    raise MdsException('Array bounds unimplemented')
            
            # The inverse from what the documentation claims...
            if not dsc.aflags.column:
                order = 'F' # Fortran-style Column-Major

            data = None
            data_buffer = buffer[ dsc.offset : dsc.offset + dsc.arsize ]
            
            if dtype_id in NUMPY_DTYPE_MAP:
                data = numpy.frombuffer(data_buffer, dtype=NUMPY_DTYPE_MAP[dtype_id], count=count)
            
            elif dtype_id == DTYPE_T:
                data = numpy.frombuffer(data_buffer, dtype=f'|S{dsc.length}').astype(str)

            elif dsc.dtype_id in [ DTYPE_F, DTYPE_D, DTYPE_G ]:
                data = convert_float_array(dsc.dtype_id, data_buffer)

            data = data.reshape(shape, order=order)
                
            return dtype_class(data)
        
        elif issubclass(dtype_class, DescriptorAPD):

            dsc = mdsdsc_a_t.from_buffer(buffer)

            if dsc.offset == 0:
                dsc.offset = ctypes.sizeof(dsc)

            if dsc.length != 4:
                raise MdsException('APD length != 4')

            offsets_buffer = buffer[ dsc.offset : dsc.offset + dsc.arsize ]
            offsets = numpy.frombuffer(offsets_buffer, dtype=numpy.uint32)

            descs = []
            for offset in offsets:
                if offset == 0:
                    descs.append(Descriptor())
                    continue

                data_buffer = buffer[offset : ]
                descs.append(Descriptor.unpack(data_buffer, conn=conn))

            return dtype_class(descs=descs)
            
        elif issubclass(dtype_class, DescriptorR):

            dsc = mdsdsc_r_t.from_buffer(buffer)

            offsets_buffer = buffer[ ctypes.sizeof(dsc) : ]
            offsets = numpy.frombuffer(offsets_buffer, dtype=numpy.uint32, count=dsc.ndesc)

            arguments = []

            data_buffer = buffer[ dsc.offset : ]
            if dsc.length == 1:
                arguments.append(Int8(numpy.frombuffer(data_buffer, dtype=numpy.uint8, count=1)[0]))
            elif dsc.length == 2:
                arguments.append(Int16(numpy.frombuffer(data_buffer, dtype=numpy.uint16, count=1)[0]))

            for i in range(dsc.ndesc):
                if offsets[i] == 0:
                    arguments.append(Descriptor())
                    continue

                dscptr_buffer = buffer[ offsets[i] : ]
                arguments.append(Descriptor.unpack(dscptr_buffer, conn=conn))

            return dtype_class(*arguments)

###
### DescriptorS
###

class DescriptorS(Descriptor):
    """
    A subclass of :class:`Descriptor` representing CLASS_S types.
    """

    def __new__(cls, *args, **kwargs):
        if cls is DescriptorS:
            raise Exception(f'DescriptorS cannot be instantiated directly')
        
        return object.__new__(cls)

    def __init__(self, data, dsc):
        
        dsc.class_id = CLASS_S
        dsc.offset = ctypes.sizeof(dsc)

        super().__init__(data=data, dsc=dsc)
    
    def pack(self):
        return self.pack_header() + self.pack_data()
    
    def pack_header(self):
        """
        Pack just the `mdsdsc_s_t` header into a bytearray.

        :return: The serialized bytes of the `mdsdsc_s_t`.
        :rtype: `bytearray`
        """

        return bytearray(self._dsc)
        
    def pack_data(self):
        """
        Pack just the data into a bytearray.

        :return: The serialized data.
        :rtype: `bytearray`
        """

        return bytearray(self._data.tobytes())
    
    @staticmethod
    def unpack_data(dtype_id, buffer, length=0):
        """
        Unpack just the data, along with a given `dtype_id` in lieu of a full `mdsdsc_s_t.
        This is overridden by some subclasses.

        :param int dtype_id: The `DTYPE_{type}` enum corresponding to the data type.
        :return: An instance of a :class:`DescriptorS` subclass containing the data.
        :rtype: A subclass of :class:`DescriptorS`
        """

        dtype_class = DTYPE_CLASS_MAP[CLASS_S][dtype_id]

        if dtype_id in NUMPY_DTYPE_MAP:
            numpy_dtype = NUMPY_DTYPE_MAP[dtype_id]
            data = numpy.frombuffer(buffer, dtype=numpy_dtype, count=1)[0]

        elif dtype_id == DTYPE_T:
            if length > 0:
                data = buffer[ : length ].decode('ascii')
            else:
                data = buffer.decode('ascii')

        return dtype_class(data)

class String(DescriptorS):

    def __init__(self, data=''):
        data = str(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=len(data.encode('ascii')),
                dtype_id=DTYPE_T,
            ),
        )

    def __str__(self):
        return self._data

    def __repr__(self):
        return f'"{self._data}"'

    def pack_data(self):
        return bytearray(self._data.encode('ascii'))
    
    @staticmethod
    def unpack_data(buffer, length=0):
        if length > 0:
            return String(buffer[ : length ].decode('ascii'))
        
        return String(buffer.decode('ascii'))

class Ident(DescriptorS):
    def __init__(self, data=''):
        data = str(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=len(data.encode('ascii')),
                dtype_id=DTYPE_IDENT,
            ),
        )

    def __repr__(self):
        return str(self._data)

    def pack_data(self):
        return bytearray(self._data.encode('ascii'))
    
    @staticmethod
    def unpack_data(buffer, length=0):
        if length > 0:
            return Ident(buffer[ : length ].decode('ascii'))
        
        return Ident(buffer.decode('ascii'))

class TreeNID(DescriptorS):
    def __init__(self, data=0, conn=None):
        if isinstance(data, Descriptor):
            data = data.data()

        data = numpy.uint32(data)
        self._conn = conn

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_NID,
            ),
        )

    def __repr__(self):
        if self._conn is not None:
            return self._conn.get(f'getnci({self._data}, "FULLPATH")').data()
        
        return f'NID({self._data})'

class TreePath(DescriptorS):
    def __init__(self, data=''):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = str(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=len(data.encode('ascii')),
                dtype_id=DTYPE_T,
            ),
        )

    def __repr__(self):
        return str(self._data)

    def pack_data(self):
        return bytearray(self._data.encode('ascii'))
    
    @staticmethod
    def unpack_data(buffer, length=0):
        if length > 0:
            return TreePath(buffer[ : length ].decode('ascii'))
    
        return TreePath(buffer.decode('ascii'))

class UInt8(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()

        data = numpy.uint8(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_BU,
            ),
        )

    def __repr__(self):
        return f'{self.data()}BU'
    
class UInt16(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.uint16(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_WU,
            ),
        )

    def __repr__(self):
        return f'{self.data()}WU'

class UInt32(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.uint32(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_LU,
            ),
        )

    def __repr__(self):
        return f'{self.data()}LU'

class UInt64(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.uint64(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_QU,
            ),
        )

    def __repr__(self):
        return f'{self.data()}QU'

class Int8(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.int8(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_B,
            ),
        )

    def __repr__(self):
        return f'{self.data()}B'

class Int16(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.int16(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_W,
            ),
        )

    def __repr__(self):
        return f'{self.data()}W'

class Int32(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.int32(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_L,
            ),
        )

    def __repr__(self):
        return f'{self.data()}L'

class Int64(DescriptorS, Numeric):
    def __init__(self, data=0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.int64(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_Q,
            ),
        )

    def __repr__(self):
        return f'{self.data()}Q'

class Float32(DescriptorS, Numeric):
    def __init__(self, data=0.0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.float32(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_FS,
            ),
        )

    def __repr__(self):
        return f'{self.data()}F0'

class Float64(DescriptorS, Numeric):
    def __init__(self, data=0.0):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.float64(data)

        super().__init__(
            data=data,
            dsc=mdsdsc_s_t(
                length=data.itemsize,
                dtype_id=DTYPE_FT,
            ),
        )

    def __repr__(self):
        return f'{self.data()}D0'

###
### DescriptorA
###

class DescriptorA(Descriptor):

    def __new__(cls, *args, **kwargs):
        if cls is DescriptorA:
            raise Exception(f'DescriptorA cannot be instantiated directly')
        
        return object.__new__(cls)

    def __init__(self, data, dsc, dims=[]):
        
        self._dims = tuple(dims)

        if isinstance(data, numpy.ndarray): # TODO: StringArray
            self._dims = data.shape[::-1]
            dsc.length = data.itemsize
            dsc.arsize = data.size * data.itemsize

        dsc.offset = ctypes.sizeof(dsc)
        dsc.dimct = len(self._dims)
        dsc.aflags.redim = 1 # Indicate that this array can be redimensioned
        dsc.aflags.column = 1 # Opposite to what the comments claim, this indicates row-major
        dsc.aflags.coeff = (dsc.dimct > 1)

        if dsc.aflags.coeff:
            # a0
            dsc.offset += ctypes.sizeof(ctypes.c_uint32)

            # coeff
            dsc.offset += ctypes.sizeof(ctypes.c_uint32) * dsc.dimct

        super().__init__(data=data, dsc=dsc)

    def __eq__(self, other):
        if isinstance(other, Descriptor):
            other = other.data()

        # numpy.array(bool).all()
        return (self.data() == other).all()

    def __ne__(self, other):
        if isinstance(other, Descriptor):
            other = other.data()

        # numpy.array(bool).any()
        return (self.data() != other).any()
    
    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, index):
        return self._data[index]
    
    @property
    def scale(self):
        """mdsdsc_a_t.scale"""
        return self._dsc.scale

    @property
    def digits(self):
        """mdsdsc_a_t.digits"""
        return self._dsc.digits

    @property
    def aflags(self):
        """mdsdsc_a_t.aflags"""
        return self._dsc.aflags
    
    @property
    def dimct(self):
        """mdsdsc_a_t.dimct"""
        return self._dsc.dimct
    
    @property
    def dims(self):
        """mdsdsc_a_t.dims"""
        return self._dims
    
    @property
    def arsize(self):
        """mdsdsc_a_t.arsize"""
        return self._dsc.arsize
    
    def pack(self):
        return self.pack_header() + self.pack_data()
    
    def pack_header(self):
        buffer = bytearray(self._dsc)

        if self._dsc.aflags.coeff:
            a0_coeffs = [self._dsc.offset] + list(self._dims)
            a0_coeffs = numpy.array(a0_coeffs, dtype=numpy.int32)
            buffer += a0_coeffs.tobytes()

        return buffer
        
    def pack_data(self):
        return bytearray(self._data.tobytes())
    
    @staticmethod
    def unpack_data(dtype_id, buffer, dims=[], length=0):
        dtype_class = DTYPE_CLASS_MAP[CLASS_A][dtype_id]

        if dtype_id in NUMPY_DTYPE_MAP:
            numpy_dtype = NUMPY_DTYPE_MAP[dtype_id]
            data = numpy.frombuffer(buffer, dtype=numpy_dtype)

        elif dtype_id == DTYPE_T:
            data = numpy.frombuffer(buffer, dtype=f'|S{length}').astype(str)

        else:
            raise MdsException(f'Unable to unpack array data with {dtype_to_string(dtype_id)}')
        
        if len(dims) > 0:
            data = data.reshape(dims[::-1])

        return dtype_class(data)

class StringArray(DescriptorA):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()

        data = numpy.array(data, dtype=str)
        shape = data.shape

        # Iterating over a flat array is considerably easier than using numpy.nditer
        data = data.flatten()

        # Ensure that all elements are the same size by padding them with trailing spaces
        maxlen = data.itemsize
        for i, s in enumerate(data):
            data[i] = s.ljust(maxlen)

        # Put our original shape back
        data = data.reshape(shape)
        
        # Force numpy to use ASCII (well, UTF-8) instead of UTF-32
        data = data.astype(bytes)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_T,
            ),
        )

    def __repr__(self):
        return repr(self._data.astype(str).tolist())

    def __eq__(self, other):
        if isinstance(other, Descriptor):
            other = other.data()

        # numpy.array(bool).all()
        return (self.data().astype(str) == other.astype(str)).all()

    def __ne__(self, other):
        if isinstance(other, Descriptor):
            other = other.data()

        # numpy.array(bool).any()
        return (self.data().astype(str) != other.astype(str)).any()

    def __iter__(self):
        return self.data().__iter__()

    def __getitem__(self, index):
        return self.data()[index]
    
    def data(self):
        tmp = self._data.astype(str).flatten()

        for i, s in enumerate(tmp):
            tmp[i] = s.rstrip()

        return tmp.reshape(self._data.shape)
    
    def data_raw(self):
        return self._data.astype(str)

class UInt8Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()

        elif isinstance(data, (bytes, bytearray, memoryview)):
            data = numpy.frombuffer(data, dtype=numpy.uint8)

        else:
            data = numpy.array(data, dtype=numpy.uint8)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_BU,
            ),
        )

    def __repr__(self):
        return f'Byte_Unsigned({repr(self._data.tolist())})'

    def deserialize(self, conn=None):
        return Descriptor.unpack(bytearray(self.data().tobytes()), conn=conn)

class UInt16Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()

        data = numpy.array(data, dtype=numpy.uint16)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_WU,
            ),
        )

    def __repr__(self):
        return f'Word_Unsigned({repr(self._data.tolist())})'

class UInt32Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.uint32)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_LU,
            ),
        )

    def __repr__(self):
        return f'Long_Unsigned({repr(self._data.tolist())})'

class UInt64Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.uint64)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_QU,
            ),
        )

    def __repr__(self):
        return f'Quadword_Unsigned({repr(self._data.tolist())})'

class Int8Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        elif isinstance(data, (bytes, bytearray, memoryview)):
            data = numpy.frombuffer(data, dtype=numpy.int8)

        else:
            data = numpy.array(data, dtype=numpy.int8)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_B,
            ),
        )

    def __repr__(self):
        return f'Byte({repr(self._data.tolist())})'

    def deserialize(self, conn=None):
        return Descriptor.unpack(bytearray(self.data().tobytes()), conn=conn)

class Int16Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.int16)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_W,
            ),
        )

    def __repr__(self):
        return f'Word({repr(self._data.tolist())})'

class Int32Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.int32)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_L,
            ),
        )

    def __repr__(self):
        return f'Long({repr(self._data.tolist())})'

class Int64Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.int64)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_Q,
            ),
        )

    def __repr__(self):
        return f'Quadword({repr(self._data.tolist())})'

class Float32Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.float32)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_FS,
            ),
        )

    def __repr__(self):
        return f'FS_FLOAT({repr(self._data.tolist())})'

class Float64Array(DescriptorA, Numeric):
    def __init__(self, data=[]):
        if isinstance(data, Descriptor):
            data = data.data()
            
        data = numpy.array(data, dtype=numpy.float64)

        super().__init__(
            data=data,
            dsc=mdsdsc_a_t(
                class_id=CLASS_A,
                dtype_id=DTYPE_FT,
            ),
        )

    def __repr__(self):
        return f'FT_FLOAT({repr(self._data.tolist())})'
    
###
### DescriptorAPD
###

class DescriptorAPD(Descriptor):

    def __new__(cls, *args, **kwargs):
        if cls is DescriptorAPD:
            raise Exception(f'DescriptorAPD cannot be instantiated directly')
        
        return object.__new__(cls)

    def __init__(self, data, count, dsc):

        # TODO:
        
        dsc.class_id = CLASS_APD
        dsc.length = ctypes.sizeof(ctypes.c_uint32)
        dsc.arsize = count * dsc.length
        dsc.dimct = 1

        super().__init__(data=data, dsc=dsc)

    def data(self):
        return self

    @property
    def scale(self):
        """mdsdsc_a_t.scale"""
        return self._dsc.scale

    @property
    def digits(self):
        """mdsdsc_a_t.digits"""
        return self._dsc.digits

    @property
    def aflags(self):
        """mdsdsc_a_t.aflags"""
        return self._dsc.aflags
    
    @property
    def dimct(self):
        """mdsdsc_a_t.dimct"""
        return self._dsc.dimct
    
    @property
    def dims(self):
        """mdsdsc_a_t.1]"""
        return self._data.shape[::-1]
    
    @property
    def arsize(self):
        """mdsdsc_a_t.arsize"""
        return self._dsc.arsize
    
    @property
    def descs(self):
        return list(self)

    def __repr__(self):
        return f'{self.__class__.__name__}({",".join(map(repr, self.descs))})'
    
    def pack(self):

        count = self._dsc.arsize // self._dsc.length
        offsets = numpy.zeros(count, dtype=numpy.uint32)

        data_buffer = bytearray()
        data_offset = ctypes.sizeof(self._dsc) + (offsets.itemsize * offsets.size)

        for i, value in enumerate(self.descs):

            if type(value) is Descriptor:
                offsets[i] = 0
                continue

            offsets[i] = data_offset + len(data_buffer)
            data_buffer += value.pack()

        return bytearray(bytes(self._dsc) + offsets.tobytes() + data_buffer)
    
class List(DescriptorAPD):
    
    def __init__(self, *items, descs=[]):

        data = []

        if len(items) == 1 and isinstance(items[0], Descriptor):
            items = items[0].data()

        # If this has been called as List(list)
        if len(items) == 1 and isinstance(items[0], (list, tuple)):
            for item in items[0]:
                data.append(Descriptor.from_data(item))

        else:
            for item in items:
                data.append(Descriptor.from_data(item))

        data.extend(descs)

        super().__init__(
            data=data,
            count=len(data),
            dsc=mdsdsc_a_t(
                dtype_id=DTYPE_LIST,
            )
        )

    def __repr__(self):
        if len(self._data) == 0:
            return 'List()'
        return f'List(,{",".join(map(repr, self._data))})'

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, index):
        return self._data[index]
    
    def __setitem__(self, index, value):
        self._data[index] = Descriptor.from_data(value)
        self._dsc.arsize = len(self._data) * self._dsc.length
    
    def append(self, value):
        self._data.append(Descriptor.from_data(value))
        self._dsc.arsize = len(self._data) * self._dsc.length

    def remove(self, value):
        self._data.remove(Descriptor.from_data(value))
        self._dsc.arsize = len(self._data) * self._dsc.length

    def data(self):
        return self._data

# Trying to subclass tuple causes issues
class Tuple(DescriptorAPD):
    
    def __init__(self, *items, descs=[]):

        data = []

        if len(items) == 1 and isinstance(items[0], Descriptor):
            items = items[0].data()
            
        # If this has been called as Tuple(tuple)
        if len(items) == 1 and isinstance(items[0], (list, tuple)):
            for item in items[0]:
                data.append(Descriptor.from_data(item))

        else:
            for item in items:
                data.append(Descriptor.from_data(item))

        data.extend(descs)

        super().__init__(
            data=tuple(data),
            count=len(data),
            dsc=mdsdsc_a_t(
                dtype_id=DTYPE_TUPLE,
            )
        )

    def __repr__(self):
        if len(self._data) == 0:
            return 'Tuple()'
        return f'Tuple(,{",".join(map(repr, self._data))})'

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, index):
        return self._data[index]

    def data(self):
        return tuple(self._data)

class Dictionary(DescriptorAPD):
    
    # dict or key, value, ...repeat
    def __init__(self, *pairs, descs=[]):

        if len(pairs) == 1 and isinstance(pairs[0], Descriptor):
            pairs = pairs[0].data()

        # If this has been called as Dictionary(dict)
        if len(pairs) == 1 and isinstance(pairs[0], dict):
            pairs = list(pairs[0].items())

        # If the entire list contains key/value tuples
        elif all([ (isinstance(pair, tuple) and len(pair) == 2) for pair in pairs ]):
            pairs = list(pairs)
            
        else:
            from itertools import islice, izip
            pairs = izip(islice(pairs, None, None, 2), islice(pairs, 1, None, 2))

        data = {}
        for k, v in pairs:
            data[Descriptor.from_data(k)] = Descriptor.from_data(v)

        if (len(descs) % 2) != 0:
            raise Exception('Cannot create Dictionary, descs must be a list of key/value pairs')

        for i in range(0, len(descs), 2):
            data[descs[i]] = descs[i + 1]

        super().__init__(
            data=data,
            count=len(data) * 2,
            dsc=mdsdsc_a_t(
                dtype_id=DTYPE_DICTIONARY,
            )
        )

    def __repr__(self):
        return f'Dictionary({", ".join(list(map(repr, self.items())))})'

    def __repr__(self):
        if len(self._data) == 0:
            return 'Dict()'
        return f'Dict(,{",".join([ f" {repr(k)},{repr(v)}" for k,v in self._data.items() ])})'
    
    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()
    
    def __contains__(self, key):
        return self._data.__contains__(key)
    
    def __getitem__(self, key):
        return self._data.__getitem__(key)
    
    def __setitem__(self, key, value):
        self._data[Descriptor.from_data(key)] = Descriptor.from_data(value)
        self._dsc.arsize = (len(self._data) * 2) * self._dsc.length
    
    def keys(self):
        return self._data.keys()
    
    def values(self):
        return self._data.values()
    
    def items(self):
        return self._data.items()
    
    def get(self, key, default):
        return self._data.get(key, default)

    def data(self):
        return dict(self._data)
    
    @property
    def descs(self):
        descs = []
        for k, v in self.items():
            descs.append(k)
            descs.append(v)
        return descs

###
### DescriptorR
###

# TODO: data = Signal(); data2 = Signal(data)

class DescriptorR(Descriptor):

    def __new__(cls, *args, **kwargs):
        if cls is DescriptorR:
            raise Exception(f'DescriptorR cannot be instantiated directly')
        
        return object.__new__(cls)
    
    def __init__(self, dsc, dscptrs, data=None):
        
        dsc.class_id = CLASS_R
        dsc.ndesc = len(dscptrs)
        
        self._dscptrs = [ Descriptor.from_data(dscptr) for dscptr in dscptrs ]

        if data is not None:
            dsc.length = data.itemsize

        super().__init__(data=data, dsc=dsc)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(map(repr, self._dscptrs))})'
    
    def __eq__(self, other):
        if type(self) != type(other):
            return False
        
        if self._data != other._data:
            return False
        
        if len(self._dscptrs) != len(other._dscptrs):
            return False
        
        for ours, theirs in zip(self._dscptrs, other._dscptrs):
            if ours != theirs:
                return False

        return True
    
    @property
    def ndesc(self):
        """mdsdsc_r_t.ndesc"""
        return self._dsc.ndesc
    
    @property
    def dscptrs(self):
        """mdsdsc_r_t.dscptrs"""
        return self._dscptrs

    def data(self):
        return self
    
    def pack(self):
        offsets = numpy.zeros(self._dsc.ndesc, dtype=numpy.uint32)

        dscptrs_buffer = bytearray()
        dscptrs_offset = ctypes.sizeof(self._dsc) + (offsets.itemsize * offsets.size)

        data_buffer = bytearray()
        if self._data is not None:
            self._dsc.length = self._data.itemsize
            self._dsc.offset = dscptrs_offset
            dscptrs_offset += self._data.itemsize
            data_buffer = bytearray(self._data.tobytes())

        for i, dscptr in enumerate(self.dscptrs):

            if type(dscptr) is Descriptor:
                offsets[i] = 0
                continue

            offsets[i] = dscptrs_offset + len(dscptrs_buffer)
            dscptrs_buffer += dscptr.pack()

        return bytearray(bytes(self._dsc) + offsets.tobytes() + data_buffer + dscptrs_buffer)

class Signal(DescriptorR):
    
    def __init__(self, value=None, raw=None, *dimensions):
        super().__init__(
            dscptrs=[value, raw, *dimensions],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_SIGNAL,
            )
        )

    @property
    def value(self):
        return self._dscptrs[0]
    
    @property
    def raw(self):
        return self._dscptrs[1]
    
    @property
    def dimensions(self):
        return self._dscptrs[2 : ]

    def __repr__(self):
        return f'Build_Signal({", ".join(map(repr, self._dscptrs))})'

    def data(self):
        if type(self.value) is not Descriptor:
            return self.value.data()
        return self.raw.data()
    
    def raw_of(self):
        return self._dscptrs[1]

    def dim_of(self, index=0):
        return self.dimensions[index]

class Dimension(DescriptorR):
    
    def __init__(self, window=None, axis=None):
        super().__init__(
            dscptrs=[window, axis],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_DIMENSION,
            )
        )

    @property
    def window(self):
        return self._dscptrs[0]
    
    @property
    def axis(self):
        return self._dscptrs[1]

    def __repr__(self):
        return f'Build_Dimension({", ".join(map(repr, self._dscptrs))})'

class Window(DescriptorR):
    
    def __init__(self, startidx=None, endingidx=None, value_at_idx0=None):
        super().__init__(
            dscptrs=[startidx, endingidx, value_at_idx0],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_WINDOW,
            )
        )

    @property
    def startidx(self):
        return self._dscptrs[0]

    @property
    def endingidx(self):
        return self._dscptrs[1]

    @property
    def value_at_idx0(self):
        return self._dscptrs[2]

    def __repr__(self):
        return f'Build_Window({", ".join(map(repr, self._dscptrs))})'

class Slope(DescriptorR):
    
    # slope, begin, ending, ...repeat
    def __init__(self, *segments):
        if (len(segments) % 3) != 0:
            raise Exception('len(segments) must be a multiple of 3')
        
        super().__init__(
            dscptrs=[*segments],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_SLOPE,
            )
        )

    @property
    def segments(self):
        segments = []
        for i in range(0, len(self._dscptrs), 3):
            segments.append((self._dscptrs[i], self._dscptrs[i + 1], self._dscptrs[i + 2]))
        return segments

    def __repr__(self):
        return f'Build_Slope({", ".join(map(repr, self._dscptrs))})'

class Function(DescriptorR):
    
    def __init__(self, opcode=0, *arguments):

        # e.g. 'ADD' -> 38
        if isinstance(opcode, str):
            from .internals.opcbuiltins import OPCODE_MAP
            opcode = OPCODE_MAP.get(opcode.upper(), -1)

        super().__init__(
            data=numpy.uint16(opcode),
            dscptrs=[*arguments],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_FUNCTION,
            )
        )

    @property
    def opcode(self):
        return self._data

    @property
    def arguments(self):
        return self._dscptrs

    def __repr__(self):
        from .internals.opcbuiltins import OPCODE_REPR_MAP
        if self._data in OPCODE_REPR_MAP:
            return OPCODE_REPR_MAP[self._data](*self._dscptrs)
        return f'Build_Function({self._data}, {", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Evaluating functions is not implemented')

class Conglom(DescriptorR):
    
    def __init__(self, image=None, model=None, name=None, qualifiers=None):
        super().__init__(
            dscptrs=[image, model, name, qualifiers],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_CONGLOM,
            )
        )

    @property
    def image(self):
        return self._dscptrs[0]

    @property
    def model(self):
        return self._dscptrs[1]

    @property
    def name(self):
        return self._dscptrs[2]

    @property
    def qualifiers(self):
        return self._dscptrs[3]

    def __repr__(self):
        return f'Build_Conglom({", ".join(map(repr, self._dscptrs))})'

class Range(DescriptorR):
    
    def __init__(self, begin=None, ending=None, deltaval=None):
        super().__init__(
            dscptrs=[begin, ending, deltaval],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_RANGE,
            )
        )

    @property
    def begin(self):
        return self._dscptrs[0]

    @property
    def ending(self):
        return self._dscptrs[1]

    @property
    def deltaval(self):
        return self._dscptrs[2]

    def __repr__(self):
        return f'Build_Range({", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        return range(self.begin.data(), self.ending.data(), self.deltaval.data())
    
    # TODO: iterator?
    
class Action(DescriptorR):

    def __init__(self, dispatch=None, task=None, errorlogs=None, completion_message=None, performance=None):
        super().__init__(
            dscptrs=[dispatch, task, errorlogs, completion_message, performance],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_ACTION,
            )
        )
    
    @property
    def dispatch(self):
        return self._dscptrs[0]
    
    @property
    def task(self):
        return self._dscptrs[1]
    
    @property
    def errorlogs(self):
        return self._dscptrs[2]
    
    @property
    def completion_message(self):
        return self._dscptrs[3]
    
    @property
    def performance(self):
        return self._dscptrs[4]

    def __repr__(self):
        return f'Build_Action({", ".join(map(repr, self._dscptrs))})'

class Dispatch(DescriptorR):
    
    def __init__(self, treesched=0, ident=None, phase=None, when=None, completion=None):
        super().__init__(
            data=numpy.uint8(treesched),
            dscptrs=[ident, phase, when, completion],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_DISPATCH,
            )
        )

    @property
    def treesched(self):
        return self._data

    @property
    def ident(self):
        return self._dscptrs[0]

    @property
    def phase(self):
        return self._dscptrs[1]

    @property
    def when(self):
        return self._dscptrs[2]

    @property
    def completion(self):
        return self._dscptrs[3]

    def __repr__(self):
        return f'Build_Dispatch({self._data}, {", ".join(map(repr, self._dscptrs))})'

class Program(DescriptorR):
    
    def __init__(self, time_out=None, program=None):
        super().__init__(
            dscptrs=[time_out, program],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_PROGRAM,
            )
        )

    @property
    def time_out(self):
        return self._dscptrs[0]

    @property
    def program(self):
        return self._dscptrs[1]

    def __repr__(self):
        return f'Build_Program({", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Programs are not implemented')

class Routine(DescriptorR):
    
    def __init__(self, time_out=None, image=None, routine=None, *arguments):
        super().__init__(
            dscptrs=[time_out, image, routine, *arguments],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_ROUTINE,
            )
        )

    @property
    def time_out(self):
        return self._dscptrs[0]

    @property
    def image(self):
        return self._dscptrs[1]

    @property
    def routine(self):
        return self._dscptrs[2]

    @property
    def arguments(self):
        return self._dscptrs[3 : ]

    def __repr__(self):
        return f'Build_Routine({", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Routines are not implemented')

class Procedure(DescriptorR):
    
    def __init__(self, time_out=None, language=None, procedure=None, *arguments):
        super().__init__(
            dscptrs=[time_out, language, procedure, *arguments],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_PROCEDURE,
            )
        )

    @property
    def time_out(self):
        return self._dscptrs[0]

    @property
    def language(self):
        return self._dscptrs[1]

    @property
    def procedure(self):
        return self._dscptrs[2]

    @property
    def arguments(self):
        return self._dscptrs[3 : ]

    def __repr__(self):
        return f'Build_Procedure({", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Procedures are not implemented')

class Method(DescriptorR):
    
    def __init__(self, time_out=None, method=None, device=None):
        super().__init__(
            dscptrs=[time_out, method, device],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_METHOD,
            )
        )

    @property
    def time_out(self):
        return self._dscptrs[0]

    @property
    def method(self):
        return self._dscptrs[1]

    @property
    def device(self):
        return self._dscptrs[2]

    def __repr__(self):
        return f'Build_Method({", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Methods are not implemented')

class Dependency(DescriptorR):
    
    def __init__(self, treedep=0, *arguments):
        super().__init__(
            data=numpy.uint8(treedep),
            dscptrs=[*arguments],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_DEPENDENCY,
            )
        )

    @property
    def treedep(self):
        return self._data

    @property
    def arguments(self):
        return self._dscptrs

    def __repr__(self):
        return f'Build_Dependency({self._data}, {", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Dependencies are not implemented')

class Condition(DescriptorR):
    
    def __init__(self, treecond=0, condition=None):
        super().__init__(
            data=numpy.uint8(treecond),
            dscptrs=[condition],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_CONDITION,
            )
        )

    @property
    def treecond(self):
        return self._data

    @property
    def condition(self):
        return self._dscptrs[0]

    def __repr__(self):
        return f'Build_Condition({self._data}, {", ".join(map(repr, self._dscptrs))})'

class WithUnits(DescriptorR):
    
    def __init__(self, value=None, units=None):
        super().__init__(
            dscptrs=[value, units],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_WITH_UNITS,
            )
        )

    @property
    def value(self):
        return self._dscptrs[0]

    @property
    def units(self):
        return self._dscptrs[1]

    def __repr__(self):
        return f'Build_With_Units({", ".join(map(repr, self._dscptrs))})'

    def data(self):
        return self.value.data()

class Call(DescriptorR):
    
    def __init__(self, return_dtype_id=DTYPE_L, image=None, routine=None, *arguments):
        super().__init__(
            data=numpy.uint8(return_dtype_id),
            dscptrs=[image, routine, *arguments],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_CALL,
            )
        )

    @property
    def return_dtype_id(self):
        return self._data

    @property
    def image(self):
        return self._dscptrs[0]
    
    @property
    def routine(self):
        return self._dscptrs[1]
    
    @property
    def arguments(self):
        return self._dscptrs[2 : ]

    def __repr__(self):
        return f'Build_Call({self._data}, {", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        raise Exception('Calls are not implemented')

class WithError(DescriptorR):
    
    def __init__(self, value=None, error=None):
        super().__init__(
            dscptrs=[value, error],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_WITH_ERROR,
            )
        )

    @property
    def value(self):
        return self._dscptrs[0]

    @property
    def error(self):
        return self._dscptrs[1]

    def __repr__(self):
        return f'Build_With_Error({", ".join(map(repr, self._dscptrs))})'
    
    def getException(self):
        return getExceptionFromError(self.error.data())

    def data(self):
        return self.value.data()

class Opaque(DescriptorR):
    
    def __init__(self, value=None, opaque_type=None):
        super().__init__(
            dscptrs=[value, opaque_type],
            dsc=mdsdsc_r_t(
                dtype_id=DTYPE_OPAQUE,
            )
        )

    @property
    def value(self):
        return self._dscptrs[0]

    @property
    def opaque_type(self):
        return self._dscptrs[1]

    def __repr__(self):
        return f'Build_Opaque({", ".join(map(repr, self._dscptrs))})'
    
    def data(self):
        return self.value.data()

###
### Lookup Tables
###

CLASS_MAP = {
    CLASS_MISSING: Descriptor,
    CLASS_S: DescriptorS,
    CLASS_A: DescriptorA,
    CLASS_CA: DescriptorA,
    CLASS_APD: DescriptorA,
    CLASS_R: DescriptorR,
}

DTYPE_CLASS_MAP = {
    CLASS_S: {
        CLASS_MISSING: Descriptor,
        DTYPE_T: String,
        DTYPE_IDENT: Ident,
        DTYPE_PATH: TreePath,
        DTYPE_NID: TreeNID,
        DTYPE_BU: UInt8,
        DTYPE_WU: UInt16,
        DTYPE_LU: UInt32,
        DTYPE_QU: UInt64,
        DTYPE_B: Int8,
        DTYPE_W: Int16,
        DTYPE_L: Int32,
        DTYPE_Q: Int64,
        DTYPE_FS: Float32,
        DTYPE_F: Float32,
        DTYPE_FT: Float64,
        DTYPE_D: Float64,
        DTYPE_G: Float64,
    },
    CLASS_A: {
        CLASS_MISSING: Descriptor,
        DTYPE_T: StringArray,
        DTYPE_BU: UInt8Array,
        DTYPE_WU: UInt16Array,
        DTYPE_LU: UInt32Array,
        DTYPE_QU: UInt64Array,
        DTYPE_B: Int8Array,
        DTYPE_W: Int16Array,
        DTYPE_L: Int32Array,
        DTYPE_Q: Int64Array,
        DTYPE_FS: Float32Array,
        DTYPE_F: Float32Array,
        DTYPE_FT: Float64Array,
        DTYPE_D: Float64Array,
        DTYPE_G: Float64Array,
    },
    CLASS_APD: {
        DTYPE_LIST: List,
        DTYPE_TUPLE: Tuple,
        DTYPE_DICTIONARY: Dictionary,
    },
    CLASS_R: {
        CLASS_MISSING: Descriptor,
        DTYPE_SIGNAL: Signal,
        DTYPE_DIMENSION: Dimension,
        DTYPE_WINDOW: Window,
        DTYPE_SLOPE: Slope,
        DTYPE_FUNCTION: Function,
        DTYPE_CONGLOM: Conglom,
        DTYPE_RANGE: Range,
        DTYPE_ACTION: Action,
        DTYPE_DISPATCH: Dispatch,
        DTYPE_PROGRAM: Program,
        DTYPE_ROUTINE: Routine,
        DTYPE_PROCEDURE: Procedure,
        DTYPE_METHOD: Method,
        DTYPE_DEPENDENCY: Dependency,
        DTYPE_CONDITION: Condition,
        DTYPE_WITH_UNITS: WithUnits,
        DTYPE_CALL: Call,
        DTYPE_WITH_ERROR: WithError,
        DTYPE_OPAQUE: Opaque,
    }
}

NUMPY_DTYPE_MAP = {
    DTYPE_BU: numpy.uint8,
    DTYPE_WU: numpy.uint16,
    DTYPE_LU: numpy.uint32,
    DTYPE_QU: numpy.uint64,
    DTYPE_B: numpy.int8,
    DTYPE_W: numpy.int16,
    DTYPE_L: numpy.int32,
    DTYPE_Q: numpy.int64,
    DTYPE_FS: numpy.float32,
    DTYPE_FT: numpy.float64,
    DTYPE_NID: numpy.uint32,
}


#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import ctypes

###
### dtype_t
###

dtype_t = ctypes.c_ubyte

DTYPE_MISSING = 0
"""No data"""

DTYPE_V = 1
"""(deprecated) Aligned bit string"""

DTYPE_BU = 2
"""Byte Unsigned, 8-bit unsigned integer"""

DTYPE_WU = 3
"""Word Unsigned, 16-bit unsigned integer"""

DTYPE_LU = 4
"""Long Unsigned, 32-bit integer"""

DTYPE_QU = 5
"""Quadword Unsigned, 64-bit integer"""

DTYPE_B = 6
"""Byte, 8-bit signed integer"""

DTYPE_W = 7
"""Word, 16-bit signed integer"""

DTYPE_L = 8
"""Long 32-bit signed integer"""

DTYPE_Q = 9
"""Quadword 64-bit signed integer"""

DTYPE_F = 10
"""(deprecated) VMS F_Floating 32-bit single-precision floating point"""

DTYPE_D = 11
"""(deprecated) VMS D_Floating 64-bit single-precision floating point"""

DTYPE_FC = 12
"""(deprecated) DTYPE_F Complex"""

DTYPE_DC = 13
"""(deprecated) DTYPE_D Complex"""

DTYPE_T = 14
"""Text, ASCII 8-bit character string"""

DTYPE_DSC = 24
"""Descriptor"""

DTYPE_OU = 25
"""Octaword Unsigned, 128-bit unsigned integer"""

DTYPE_O = 26
"""Unsigned Octaword, 128-bit signed integer"""

DTYPE_G = 27
"""(deprecated) VMS G_floating 64-bit double-precision floating point"""

DTYPE_H = 28
"""(deprecated) VMS H_Floating 128-bit quadruple-precision floating point"""

DTYPE_GC = 29
"""(deprecated) DTYPE_G complex"""

DTYPE_HC = 30
"""(deprecated) DTYPE_H complex"""

DTYPE_POINTER = 51
"""TODO: """

DTYPE_FS = 52
"""IEEE 32-bit single-precision floating point"""

DTYPE_FT = 53
"""IEEE 64-bit double-precision floating point"""

DTYPE_FSC = 54
"""DTYPE_FS complex"""

DTYPE_FTC = 55
"""DTYPE_FT complex"""

DTYPE_C = 56
"""Used for char* returned by TdiCall() that must be freed"""

DTYPE_IDENT = 191
"""TODO: """

DTYPE_NID = 192
"""Node identification number, see NID"""

DTYPE_PATH = 193
"""Path to node as a string"""

DTYPE_PARAM = 194
"""TODO: """

DTYPE_SIGNAL = 195
"""Signal()"""

DTYPE_DIMENSION = 196
"""Dimension()"""

DTYPE_WINDOW = 197
"""Window()"""

DTYPE_SLOPE = 198
"""Slope()"""

DTYPE_FUNCTION = 199
"""TODO: """

DTYPE_CONGLOM = 200
"""TODO: """

DTYPE_RANGE = 201
"""Range()"""

DTYPE_ACTION = 202
"""Action()"""

DTYPE_DISPATCH = 203
"""Dispatch()"""

DTYPE_PROGRAM = 204
"""(deprecated) TODO: """

DTYPE_ROUTINE = 205
"""(deprecated) TODO: """

DTYPE_PROCEDURE = 206
"""(deprecated) TODO: """

DTYPE_METHOD = 207
"""Method()"""

DTYPE_DEPENDENCY = 208
"""TODO: """

DTYPE_CONDITION = 209
"""TODO: """

DTYPE_EVENT = 210
"""TODO: """

DTYPE_WITH_UNITS = 211
"WithUnits()"

DTYPE_CALL = 212
"""TODO: """

DTYPE_WITH_ERROR = 213
"""WithError()"""

DTYPE_LIST = 214
"""List()"""

DTYPE_TUPLE = 215
"""TODO: """

DTYPE_DICTIONARY = 216
"""TODO: """

DTYPE_OPAQUE = 217
"""TODO: """

NUMERIC_DTYPES = [
    DTYPE_BU, DTYPE_WU, DTYPE_LU, DTYPE_QU,
    DTYPE_B, DTYPE_W, DTYPE_L, DTYPE_Q,
    DTYPE_F, DTYPE_D, DTYPE_FC, DTYPE_DC,
    DTYPE_OU, DTYPE_O,
    DTYPE_G, DTYPE_H, DTYPE_GC, DTYPE_HC,
    DTYPE_FS, DTYPE_FT, DTYPE_FSC, DTYPE_FTC,
    DTYPE_DSC,
]

def dtype_is_numeric(dtype_id) -> bool:
    if type(dtype_id) is dtype_t:
        dtype_id = dtype_id.value

    return dtype_id in NUMERIC_DTYPES

EXPRESSION_DTYPES = [
    DTYPE_FUNCTION,
    DTYPE_NID,
    DTYPE_PATH,
    DTYPE_IDENT,
    DTYPE_CALL
]

def dtype_is_expression(dtype_id) -> bool:
    if type(dtype_id) is dtype_t:
        dtype_id = dtype_id.value

    return dtype_id in EXPRESSION_DTYPES

def dtype_to_string(dtype_id) -> str:
    VALUES = {
        DTYPE_MISSING: 'DTYPE_MISSING',
        DTYPE_V: 'DTYPE_V',
        DTYPE_BU: 'DTYPE_BU',
        DTYPE_WU: 'DTYPE_WU',
        DTYPE_LU: 'DTYPE_LU',
        DTYPE_QU: 'DTYPE_QU',
        DTYPE_B: 'DTYPE_B',
        DTYPE_W: 'DTYPE_W',
        DTYPE_L: 'DTYPE_L',
        DTYPE_Q: 'DTYPE_Q',
        DTYPE_F: 'DTYPE_F',
        DTYPE_D: 'DTYPE_D',
        DTYPE_FC: 'DTYPE_FC',
        DTYPE_DC: 'DTYPE_DC',
        DTYPE_T: 'DTYPE_T',
        DTYPE_DSC: 'DTYPE_DSC',
        DTYPE_OU: 'DTYPE_OU',
        DTYPE_O: 'DTYPE_O',
        DTYPE_G: 'DTYPE_G',
        DTYPE_H: 'DTYPE_H',
        DTYPE_GC: 'DTYPE_GC',
        DTYPE_HC: 'DTYPE_HC',
        DTYPE_POINTER: 'DTYPE_POINTER',
        DTYPE_FS: 'DTYPE_FS',
        DTYPE_FT: 'DTYPE_FT',
        DTYPE_FSC: 'DTYPE_FSC',
        DTYPE_FTC: 'DTYPE_FTC',
        DTYPE_C: 'DTYPE_C',
        DTYPE_IDENT: 'DTYPE_IDENT',
        DTYPE_NID: 'DTYPE_NID',
        DTYPE_PATH: 'DTYPE_PATH',
        DTYPE_PARAM: 'DTYPE_PARAM',
        DTYPE_SIGNAL: 'DTYPE_SIGNAL',
        DTYPE_DIMENSION: 'DTYPE_DIMENSION',
        DTYPE_WINDOW: 'DTYPE_WINDOW',
        DTYPE_SLOPE: 'DTYPE_SLOPE',
        DTYPE_FUNCTION: 'DTYPE_FUNCTION',
        DTYPE_CONGLOM: 'DTYPE_CONGLOM',
        DTYPE_RANGE: 'DTYPE_RANGE',
        DTYPE_ACTION: 'DTYPE_ACTION',
        DTYPE_DISPATCH: 'DTYPE_DISPATCH',
        DTYPE_PROGRAM: 'DTYPE_PROGRAM',
        DTYPE_ROUTINE: 'DTYPE_ROUTINE',
        DTYPE_PROCEDURE: 'DTYPE_PROCEDURE',
        DTYPE_METHOD: 'DTYPE_METHOD',
        DTYPE_DEPENDENCY: 'DTYPE_DEPENDENCY',
        DTYPE_CONDITION: 'DTYPE_CONDITION',
        DTYPE_EVENT: 'DTYPE_EVENT',
        DTYPE_WITH_UNITS: 'DTYPE_WITH_UNITS',
        DTYPE_CALL: 'DTYPE_CALL',
        DTYPE_WITH_ERROR: 'DTYPE_WITH_ERROR',
        DTYPE_LIST: 'DTYPE_LIST',
        DTYPE_TUPLE: 'DTYPE_TUPLE',
        DTYPE_DICTIONARY: 'DTYPE_DICTIONARY',
        DTYPE_OPAQUE: 'DTYPE_OPAQUE',
    }

    if type(dtype_id) is dtype_t:
        dtype_id = dtype_id.value

    if dtype_id in VALUES:
        return VALUES[dtype_id]

    return f'DTYPE_UNKNOWN({dtype_id})'

def get_dtype_size(dtype_id) -> int:
    if type(dtype_id) is dtype_t:
        dtype_id = dtype_id.value

    if dtype_id in [ DTYPE_BU, DTYPE_B ]:
        return 1
    if dtype_id in [ DTYPE_WU, DTYPE_W ]:
        return 2
    if dtype_id in [ DTYPE_LU, DTYPE_L, DTYPE_F, DTYPE_FS ]:
        return 4
    if dtype_id in [ DTYPE_QU, DTYPE_Q, DTYPE_D, DTYPE_G, DTYPE_FT ]:
        return 8
    if dtype_id in [ DTYPE_OU, DTYPE_O ]:
        return 16
    
    return 0


#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import numpy
import unittest

from ..connection import *
from ..functions import *

class WriteTest(unittest.TestCase):

    ENABLED = False
    PORT = 12345

    @classmethod
    def setUpClass(cls):
        import atexit
        import shutil
        import tempfile
        import subprocess

        if cls.ENABLED:

            # Attempt to find the mdsip executable
            mdsip = shutil.which('mdsip')
            if mdsip is None:
                cls.ENABLED = False
                return

            # Generate a temporary directory for our config and tree files
            cls.tempdir = tempfile.mkdtemp('-mdsthin-write-test')

            # We need a valid mdsip.hosts in order to spawn a mdsip server
            cls.mdsip_hosts_filename = os.path.join(cls.tempdir, 'mdsip.hosts')
            with open(cls.mdsip_hosts_filename, 'wt') as file:
                file.write('* | MAP_TO_LOCAL\n')

            # Inform the process to put tree files in the temporary directory
            mdsip_env = dict(os.environ)
            mdsip_env['thintest_path'] = cls.tempdir

            # Spawn the mdsip server
            cls.mdsip_process = subprocess.Popen(
                [
                    mdsip,
                    '-h', cls.mdsip_hosts_filename,
                    '-p', str(cls.PORT),
                    '-c', '0',
                    '-s'
                ],
                env=mdsip_env,
                # Disable output from the mdsip server, comment out to debug
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

            # Give the process time to start
            time.sleep(0.1)

            # Ensure that we kill the process even if we don't gracefully shutdown
            atexit.register(subprocess.Popen.terminate, cls.mdsip_process)

            # Setup a new connection to the server we spawned
            cls.mdsip_connection_url = f'localhost:{cls.PORT}'
            cls.conn = Connection(cls.mdsip_connection_url)

            # Create our tree and write the tree files to disk
            cls.conn.tcl("edit thintest /new")
            cls.conn.tcl("add node num /usage=numeric")
            cls.conn.tcl("add node str /usage=text")
            cls.conn.tcl("add node sig /usage=signal")
            cls.conn.tcl("write")
            cls.conn.tcl("close")
            cls.conn.tcl("set tree thintest /shot=-1")
            cls.conn.tcl("create pulse 1") # The readonly tree

        super().setUpClass()

    @classmethod
    def tearDownClass(cls):

        if cls.ENABLED:
            # Close the connection
            cls.conn.disconnect()

            # Kill the mdsip server (this could be handled by atexit, but good to stop it before trying to delete the files)
            cls.mdsip_process.terminate()

            # Cleanup the config and tree files
            files = [
                cls.mdsip_hosts_filename,
                os.path.join(cls.tempdir, 'thintest_model.tree'),
                os.path.join(cls.tempdir, 'thintest_model.datafile'),
                os.path.join(cls.tempdir, 'thintest_model.characteristics'),
                os.path.join(cls.tempdir, 'thintest_001.tree'),
                os.path.join(cls.tempdir, 'thintest_001.datafile'),
                os.path.join(cls.tempdir, 'thintest_001.characteristics'),
            ]

            for file in files:
                try:
                    os.remove(file)
                except FileNotFoundError:
                    pass

            # Cleanup the temporary directory
            os.rmdir(cls.tempdir)

        return super().tearDownClass()

    def setUp(self):
        if not self.ENABLED:
            raise unittest.SkipTest('--write was not specified or it was unable to find mdsip executable')
        
        # Open a tree for the tests to use
        self.conn.openTree('thintest', -1)

    def test_numeric(self):

        INTEGER_ARRAY = [ [2, 4], [6, 8], [16, 32] ]
        FLOAT_ARRAY = [ [1.0, 1.5], [2.0, 2.5], [3.0, 3.5] ]

        tests = [
            { 'value': UInt8(42) },
            { 'value': UInt16(42) },
            { 'value': UInt32(42) },
            { 'value': UInt64(42) },
            { 'value': Int8(42) },
            { 'value': Int16(42) },
            { 'value': Int32(42) },
            { 'value': Int64(42) },
            { 'value': Float32(42.0) },
            { 'value': Float64(42.0) },
            { 'value': UInt8Array(INTEGER_ARRAY) },
            { 'value': UInt16Array(INTEGER_ARRAY) },
            { 'value': UInt32Array(INTEGER_ARRAY) },
            { 'value': UInt64Array(INTEGER_ARRAY) },
            { 'value': Int8Array(INTEGER_ARRAY) },
            { 'value': Int16Array(INTEGER_ARRAY) },
            { 'value': Int32Array(INTEGER_ARRAY) },
            { 'value': Int64Array(INTEGER_ARRAY) },
            { 'value': Float32Array(FLOAT_ARRAY) },
            { 'value': Float64Array(FLOAT_ARRAY) },
        ]

        for info in tests:
            self.conn.put('num', '$', info['value'])
            self.assertEqual(self.conn.get('num'), info['value'])

            # Reset the value each loop
            self.conn.put('num', '*')

        self.conn.put('num', '12345Q')
        self.assertEqual(self.conn.get('num'), Int64(12345))

        self.conn.put('num', '[1Q, 2Q, 3Q, 4Q]')
        self.assertEqual(self.conn.get('num'), Int64Array([ 1, 2, 3, 4 ]))
    
    def test_text(self):

        tests = [
            { 'value': String("Hello, World!") },
            { 'value': StringArray([ "one", "seven", "thirteen" ]) },
        ]

        for info in tests:
            self.conn.put('str', '$', info['value'])
            self.assertEqual(self.conn.get('str'), info['value'])

            # Reset the value each loop
            self.conn.put('str', '*')

        self.conn.put('str', '"Hello, World!"')
        self.assertEqual(self.conn.get('str'), String("Hello, World!"))

        self.conn.put('str', '["one", "seven", "thirteen"]')
        self.assertEqual(self.conn.get('str'), StringArray([ "one", "seven", "thirteen" ]))
    
    def test_signal(self):

        data = numpy.array([ 1, 2, 3, 4])
        times = numpy.array([ 0.1, 0.2, 0.3, 0.4 ])
        signal = Signal(data, None, times)

        self.conn.put('sig', '`SerializeIn($)', signal.serialize())
        self.assertEqual(self.conn.getObject('sig'), signal)

        # Reset the value
        self.conn.put('sig', '*')

        self.conn.put('sig', 'BUILD_SIGNAL([1Q, 2Q, 3Q, 4Q], *, [0.1D0, 0.2D0, 0.3D0, 0.4D0])')
        self.assertEqual(self.conn.getObject('sig'), signal)

        # Reset the value
        self.conn.put('sig', '*')

        self.conn.put('sig', '`SerializeIn($)', BUILD_SIGNAL(data, None, times).serialize())
        self.assertEqual(self.conn.getObject('sig'), signal)

    def test_putmany(self):
        if self.conn.getServerVersion() < (7, 145, 7):
            raise unittest.SkipTest('Disabled for MDSplus < 7.145.7')

        try:
            pm = self.conn.putMany()
            pm.append('num', '$', Int32(42))
            pm.append('str', '$', String("Hello, World!"))
            pm.append('sig', '`SerializeIn($)', Signal(1, 2, 3).serialize())
            pm.execute()

            self.assertEqual(self.conn.get('num'), Int32(42))
            self.assertEqual(self.conn.get('str'), String("Hello, World!"))
            self.assertEqual(self.conn.getObject('sig'), Signal(1, 2, 3))

        except LibKEYNOTFOU:
            raise unittest.SkipTest('Disabled when PutManyExecute is missing from libMdsObjectsCppShr')

    def test_permissions(self):
        import platform

        if platform.system() == 'Windows':
            raise unittest.SkipTest('Disabled on Windows')
        
        # Change all the readonly tree files to be readonly
        filenames = [
            os.path.join(self.tempdir, 'thintest_001.tree'),
            os.path.join(self.tempdir, 'thintest_001.datafile'),
            os.path.join(self.tempdir, 'thintest_001.characteristics'),
        ]
        for filename in filenames:
            os.chmod(filename, 0o400)

        # Open the readonly tree
        self.conn.openTree('thintest', 1)

        self.assertRaises(TreeFAILURE, self.conn.put, 'num', '42')

        self.conn.closeTree('thintest', 1)


#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import ctypes

from .dtypedef import *
from .classdef import *

###
### mdsdsc_t
###

class mdsdsc_t(ctypes.LittleEndianStructure):
    """
    Descriptor base class
    """

    _pack_ = 1
    _fields_ = [
        ('length', ctypes.c_uint16),
        ('dtype_id', dtype_t),
        ('class_id', class_t),
        ('offset', ctypes.c_uint32), # pointer
    ]

    @property
    def class_str(self):
        return class_to_string(self.class_id)
    
    @property
    def dtype_str(self):
        return dtype_to_string(self.dtype_id)

###
### mdsdsc_s_t
###

class mdsdsc_s_t(mdsdsc_t):
    """
    Static fixed-length descriptor (CLASS_S)
    """

###
### mdsdsc_d_t
###

class mdsdsc_d_t(mdsdsc_t):
    """
    Dynamic string descriptor (CLASS_D)
    """

###
### mdsdsc_a_t
###

class aflags_t(ctypes.LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('fill', ctypes.c_uint8, 3),
        ('binscale', ctypes.c_uint8, 1),
        ('redim', ctypes.c_uint8, 1),
        ('column', ctypes.c_uint8, 1),
        ('coeff', ctypes.c_uint8, 1),
        ('bounds', ctypes.c_uint8, 1),
    ]

class mdsdsc_a_t(mdsdsc_t):
    """
    Array descriptor (CLASS_A, CLASS_CA, CLASS_APD)
    """

    _pack_ = 1
    _fields_ = [
        ('scale', ctypes.c_int8),
        ('digits', ctypes.c_uint8),
        ('aflags', aflags_t),
        ('dimct', ctypes.c_uint8),
        ('arsize', ctypes.c_uint32),
    ]

###
### mdsdsc_r_t
###

class mdsdsc_r_t(mdsdsc_t):
    """
    Record descriptor (CLASS_R)
    """
    
    _pack_ = 1
    _fields_ = [
        ('ndesc', ctypes.c_uint32, 8),
        ('fill', ctypes.c_uint32, 24),
    ]



#
# Copyright (c) 2024, Massachusetts Institute of Technology All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice, this
# list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from .descriptors import Function

def d():
    return Function(0)

def dA0():
    return Function(1)

def dALPHA():
    return Function(2)

def dAMU():
    return Function(3)

def dC():
    return Function(4)

def dCAL():
    return Function(5)

def dDEGREE():
    return Function(6)

def dEV():
    return Function(7)

def dFALSE():
    return Function(8)

def dFARADAY():
    return Function(9)

def dG():
    return Function(10)

def dGAS():
    return Function(11)

def dH():
    return Function(12)

def dHBAR():
    return Function(13)

def dI():
    return Function(14)

def dK():
    return Function(15)

def dME():
    return Function(16)

def dMISSING():
    return Function(17)

def dMP():
    return Function(18)

def dN0():
    return Function(19)

def dNA():
    return Function(20)

def dP0():
    return Function(21)

def dPI():
    return Function(22)

def dQE():
    return Function(23)

def dRE():
    return Function(24)

def dROPRAND():
    return Function(25)

def dRYDBERG():
    return Function(26)

def dT0():
    return Function(27)

def dTORR():
    return Function(28)

def dTRUE():
    return Function(29)

def dVALUE():
    return Function(30)

def ABORT(*args):
    return Function(31, *args)

def ABS(*args):
    return Function(32, *args)

def ABS1(*args):
    return Function(33, *args)

def ABSSQ(*args):
    return Function(34, *args)

def ACHAR(*args):
    return Function(35, *args)

def ACOS(*args):
    return Function(36, *args)

def ACOSD(*args):
    return Function(37, *args)

def ADD(*args):
    return Function(38, *args)

def ADJUSTL(*args):
    return Function(39, *args)

def ADJUSTR(*args):
    return Function(40, *args)

def AIMAG(*args):
    return Function(41, *args)

def AINT(*args):
    return Function(42, *args)

def ALL(*args):
    return Function(43, *args)

def ALLOCATED(*args):
    return Function(44, *args)

def AND(*args):
    return Function(45, *args)

def AND_NOT(*args):
    return Function(46, *args)

def ANINT(*args):
    return Function(47, *args)

def ANY(*args):
    return Function(48, *args)

def ARG(*args):
    return Function(49, *args)

def ARGD(*args):
    return Function(50, *args)

def ARG_OF(*args):
    return Function(51, *args)

def ARRAY(*args):
    return Function(52, *args)

def ASIN(*args):
    return Function(53, *args)

def ASIND(*args):
    return Function(54, *args)

def AS_IS(*args):
    return Function(55, *args)

def ATAN(*args):
    return Function(56, *args)

def ATAN2(*args):
    return Function(57, *args)

def ATAN2D(*args):
    return Function(58, *args)

def ATAND(*args):
    return Function(59, *args)

def ATANH(*args):
    return Function(60, *args)

def AXIS_OF(*args):
    return Function(61, *args)

def BACKSPACE(*args):
    return Function(62, *args)

def IBCLR(*args):
    return Function(63, *args)

def BEGIN_OF(*args):
    return Function(64, *args)

def IBITS(*args):
    return Function(65, *args)

def BREAK():
    return Function(66)

def BSEARCH(*args):
    return Function(67, *args)

def IBSET(*args):
    return Function(68, *args)

def BTEST(*args):
    return Function(69, *args)

def BUILD_ACTION(*args):
    return Function(70, *args)

def BUILD_CONDITION(*args):
    return Function(71, *args)

def BUILD_CONGLOM(*args):
    return Function(72, *args)

def BUILD_DEPENDENCY(*args):
    return Function(73, *args)

def BUILD_DIM(*args):
    return Function(74, *args)

def BUILD_DISPATCH(*args):
    return Function(75, *args)

def BUILD_EVENT(*args):
    return Function(76, *args)

def BUILD_FUNCTION(*args):
    return Function(77, *args)

def BUILD_METHOD(*args):
    return Function(78, *args)

def BUILD_PARAM(*args):
    return Function(79, *args)

def BUILD_PATH(*args):
    return Function(80, *args)

def BUILD_PROCEDURE(*args):
    return Function(81, *args)

def BUILD_PROGRAM(*args):
    return Function(82, *args)

def BUILD_RANGE(*args):
    return Function(83, *args)

def BUILD_ROUTINE(*args):
    return Function(84, *args)

def BUILD_SIGNAL(*args):
    return Function(85, *args)

def BUILD_SLOPE(*args):
    return Function(86, *args)

def BUILD_WINDOW(*args):
    return Function(87, *args)

def BUILD_WITH_UNITS(*args):
    return Function(88, *args)

def BUILTIN_OPCODE(*args):
    return Function(89, *args)

def BYTE(*args):
    return Function(90, *args)

def BYTE_UNSIGNED(*args):
    return Function(91, *args)

def CASE(*args):
    return Function(92, *args)

def CEILING(*args):
    return Function(93, *args)

def CHAR(*args):
    return Function(94, *args)

def CLASS(*args):
    return Function(95, *args)

def FCLOSE(*args):
    return Function(96, *args)

def CMPLX(*args):
    return Function(97, *args)

def COMMA(*args):
    return Function(98, *args)

def COMPILE(*args):
    return Function(99, *args)

def COMPLETION_OF(*args):
    return Function(100, *args)

def CONCAT(*args):
    return Function(101, *args)

def CONDITIONAL(*args):
    return Function(102, *args)

def CONJG(*args):
    return Function(103, *args)

def CONTINUE():
    return Function(104)

def CONVOLVE(*args):
    return Function(105, *args)

def COS(*args):
    return Function(106, *args)

def COSD(*args):
    return Function(107, *args)

def COSH(*args):
    return Function(108, *args)

def COUNT(*args):
    return Function(109, *args)

def CSHIFT(*args):
    return Function(110, *args)

def CVT(*args):
    return Function(111, *args)

def DATA(*args):
    return Function(112, *args)

def DATE_AND_TIME(*args):
    return Function(113, *args)

def DATE_TIME(*args):
    return Function(114, *args)

def DBLE(*args):
    return Function(115, *args)

def DEALLOCATE(*args):
    return Function(116, *args)

def DEBUG(*args):
    return Function(117, *args)

def DECODE(*args):
    return Function(118, *args)

def DECOMPILE(*args):
    return Function(119, *args)

def DECOMPRESS(*args):
    return Function(120, *args)

def DEFAULT(*args):
    return Function(121, *args)

def DERIVATIVE(*args):
    return Function(122, *args)

def DESCR(*args):
    return Function(123, *args)

def DIAGONAL(*args):
    return Function(124, *args)

def DIGITS(*args):
    return Function(125, *args)

def DIM(*args):
    return Function(126, *args)

def DIM_OF(*args):
    return Function(127, *args)

def DISPATCH_OF(*args):
    return Function(128, *args)

def DIVIDE(*args):
    return Function(129, *args)

def LBOUND(*args):
    return Function(130, *args)

def DO(*args):
    return Function(131, *args)

def DOT_PRODUCT(*args):
    return Function(132, *args)

def DPROD(*args):
    return Function(133, *args)

def DSCPTR(*args):
    return Function(134, *args)

def SHAPE(*args):
    return Function(135, *args)

def SIZE(*args):
    return Function(136, *args)

def KIND(*args):
    return Function(137, *args)

def UBOUND(*args):
    return Function(138, *args)

def D_COMPLEX(*args):
    return Function(139, *args)

def D_FLOAT(*args):
    return Function(140, *args)

def RANGE(*args):
    return Function(141, *args)

def PRECISION(*args):
    return Function(142, *args)

def ELBOUND(*args):
    return Function(143, *args)

def ELSE():
    return Function(144)

def ELSEWHERE():
    return Function(145)

def ENCODE(*args):
    return Function(146, *args)

def ENDFILE(*args):
    return Function(147, *args)

def END_OF(*args):
    return Function(148, *args)

def EOSHIFT(*args):
    return Function(149, *args)

def EPSILON(*args):
    return Function(150, *args)

def EQ(*args):
    return Function(151, *args)

def EQUALS(*args):
    return Function(152, *args)

def EQUALS_FIRST(*args):
    return Function(153, *args)

def EQV(*args):
    return Function(154, *args)

def ESHAPE(*args):
    return Function(155, *args)

def ESIZE(*args):
    return Function(156, *args)

def EUBOUND(*args):
    return Function(157, *args)

def EVALUATE(*args):
    return Function(158, *args)

def EXECUTE(*args):
    return Function(159, *args)

def EXP(*args):
    return Function(160, *args)

def EXPONENT(*args):
    return Function(161, *args)

def EXT_FUNCTION(*args):
    return Function(162, *args)

def FFT(*args):
    return Function(163, *args)

def FIRSTLOC(*args):
    return Function(164, *args)

def FIT(*args):
    return Function(165, *args)

def FIX_ROPRAND(*args):
    return Function(166, *args)

def FLOAT(*args):
    return Function(167, *args)

def FLOOR(*args):
    return Function(168, *args)

def FOR(*args):
    return Function(169, *args)

def FRACTION(*args):
    return Function(170, *args)

def FUN(*args):
    return Function(171, *args)

def F_COMPLEX(*args):
    return Function(172, *args)

def F_FLOAT(*args):
    return Function(173, *args)

def GE(*args):
    return Function(174, *args)

def GETNCI(*args):
    return Function(175, *args)

def GOTO(*args):
    return Function(176, *args)

def GT(*args):
    return Function(177, *args)

def G_COMPLEX(*args):
    return Function(178, *args)

def G_FLOAT(*args):
    return Function(179, *args)

def HELP_OF(*args):
    return Function(180, *args)

def HUGE(*args):
    return Function(181, *args)

def H_COMPLEX(*args):
    return Function(182, *args)

def H_FLOAT(*args):
    return Function(183, *args)

def IACHAR(*args):
    return Function(184, *args)

def IAND(*args):
    return Function(185, *args)

def IAND_NOT(*args):
    return Function(186, *args)

def ICHAR(*args):
    return Function(187, *args)

def IDENT_OF(*args):
    return Function(188, *args)

def IF(*args):
    return Function(189, *args)

def IF_ERROR(*args):
    return Function(190, *args)

def IMAGE_OF(*args):
    return Function(191, *args)

def IN(*args):
    return Function(192, *args)

def INAND(*args):
    return Function(193, *args)

def INAND_NOT(*args):
    return Function(194, *args)

def INDEX(*args):
    return Function(195, *args)

def INOR(*args):
    return Function(196, *args)

def INOR_NOT(*args):
    return Function(197, *args)

def INOT(*args):
    return Function(198, *args)

def INOUT(*args):
    return Function(199, *args)

def INQUIRE(*args):
    return Function(200, *args)

def INT(*args):
    return Function(201, *args)

def INTEGRAL(*args):
    return Function(202, *args)

def INTERPOL(*args):
    return Function(203, *args)

def INTERSECT(*args):
    return Function(204, *args)

def INT_UNSIGNED(*args):
    return Function(205, *args)

def INVERSE(*args):
    return Function(206, *args)

def IOR(*args):
    return Function(207, *args)

def IOR_NOT(*args):
    return Function(208, *args)

def IS_IN(*args):
    return Function(209, *args)

def IEOR(*args):
    return Function(210, *args)

def IEOR_NOT(*args):
    return Function(211, *args)

def LABEL(*args):
    return Function(212, *args)

def LAMINATE(*args):
    return Function(213, *args)

def LANGUAGE_OF(*args):
    return Function(214, *args)

def LASTLOC(*args):
    return Function(215, *args)

def LE(*args):
    return Function(216, *args)

def LEN(*args):
    return Function(217, *args)

def LEN_TRIM(*args):
    return Function(218, *args)

def LGE(*args):
    return Function(219, *args)

def LGT(*args):
    return Function(220, *args)

def LLE(*args):
    return Function(221, *args)

def LLT(*args):
    return Function(222, *args)

def LOG(*args):
    return Function(223, *args)

def LOG10(*args):
    return Function(224, *args)

def LOG2(*args):
    return Function(225, *args)

def LOGICAL(*args):
    return Function(226, *args)

def LONG(*args):
    return Function(227, *args)

def LONG_UNSIGNED(*args):
    return Function(228, *args)

def LT(*args):
    return Function(229, *args)

def MATMUL(*args):
    return Function(230, *args)

def MAT_ROT(*args):
    return Function(231, *args)

def MAT_ROT_INT(*args):
    return Function(232, *args)

def MAX(*args):
    return Function(233, *args)

def MAXEXPONENT(*args):
    return Function(234, *args)

def MAXLOC(*args):
    return Function(235, *args)

def MAXVAL(*args):
    return Function(236, *args)

def MEAN(*args):
    return Function(237, *args)

def MEDIAN(*args):
    return Function(238, *args)

def MERGE(*args):
    return Function(239, *args)

def METHOD_OF(*args):
    return Function(240, *args)

def MIN(*args):
    return Function(241, *args)

def MINEXPONENT(*args):
    return Function(242, *args)

def MINLOC(*args):
    return Function(243, *args)

def MINVAL(*args):
    return Function(244, *args)

def MOD(*args):
    return Function(245, *args)

def MODEL_OF(*args):
    return Function(246, *args)

def MULTIPLY(*args):
    return Function(247, *args)

def NAME_OF(*args):
    return Function(248, *args)

def NAND(*args):
    return Function(249, *args)

def NAND_NOT(*args):
    return Function(250, *args)

def NDESC(*args):
    return Function(251, *args)

def NE(*args):
    return Function(252, *args)

def NEAREST(*args):
    return Function(253, *args)

def NEQV(*args):
    return Function(254, *args)

def NINT(*args):
    return Function(255, *args)

def NOR(*args):
    return Function(256, *args)

def NOR_NOT(*args):
    return Function(257, *args)

def NOT(*args):
    return Function(258, *args)

def OBJECT_OF(*args):
    return Function(259, *args)

def OCTAWORD(*args):
    return Function(260, *args)

def OCTAWORD_UNSIGNED(*args):
    return Function(261, *args)

def ON_ERROR(*args):
    return Function(262, *args)

def OPCODE_BUILTIN(*args):
    return Function(263, *args)

def OPCODE_STRING(*args):
    return Function(264, *args)

def FOPEN(*args):
    return Function(265, *args)

def OPTIONAL(*args):
    return Function(266, *args)

def OR(*args):
    return Function(267, *args)

def OR_NOT(*args):
    return Function(268, *args)

def OUT(*args):
    return Function(269, *args)

def PACK(*args):
    return Function(270, *args)

def PHASE_OF(*args):
    return Function(271, *args)

def POST_DEC(*args):
    return Function(272, *args)

def POST_INC(*args):
    return Function(273, *args)

def POWER(*args):
    return Function(274, *args)

def PRESENT(*args):
    return Function(275, *args)

def PRE_DEC(*args):
    return Function(276, *args)

def PRE_INC(*args):
    return Function(277, *args)

def PRIVATE(*args):
    return Function(278, *args)

def PROCEDURE_OF(*args):
    return Function(279, *args)

def PRODUCT(*args):
    return Function(280, *args)

def PROGRAM_OF(*args):
    return Function(281, *args)

def PROJECT(*args):
    return Function(282, *args)

def PROMOTE(*args):
    return Function(283, *args)

def PUBLIC(*args):
    return Function(284, *args)

def QUADWORD(*args):
    return Function(285, *args)

def QUADWORD_UNSIGNED(*args):
    return Function(286, *args)

def QUALIFIERS_OF(*args):
    return Function(287, *args)

def RADIX(*args):
    return Function(288, *args)

def RAMP(*args):
    return Function(289, *args)

def RANDOM(*args):
    return Function(290, *args)

def RANDOM_SEED(*args):
    return Function(291, *args)

def DTYPE_RANGE(*args):
    return Function(292, *args)

def RANK(*args):
    return Function(293, *args)

def RAW_OF(*args):
    return Function(294, *args)

def READ(*args):
    return Function(295, *args)

def REAL(*args):
    return Function(296, *args)

def REBIN(*args):
    return Function(297, *args)

def REF(*args):
    return Function(298, *args)

def REPEAT(*args):
    return Function(299, *args)

def REPLICATE(*args):
    return Function(300, *args)

def RESHAPE(*args):
    return Function(301, *args)

def RETURN(*args):
    return Function(302, *args)

def REWIND(*args):
    return Function(303, *args)

def RMS(*args):
    return Function(304, *args)

def ROUTINE_OF(*args):
    return Function(305, *args)

def RRSPACING(*args):
    return Function(306, *args)

def SCALE(*args):
    return Function(307, *args)

def SCAN(*args):
    return Function(308, *args)

def FSEEK(*args):
    return Function(309, *args)

def SET_EXPONENT(*args):
    return Function(310, *args)

def SET_RANGE(*args):
    return Function(311, *args)

def ISHFT(*args):
    return Function(312, *args)

def ISHFTC(*args):
    return Function(313, *args)

def SHIFT_LEFT(*args):
    return Function(314, *args)

def SHIFT_RIGHT(*args):
    return Function(315, *args)

def SIGN(*args):
    return Function(316, *args)

def SIGNED(*args):
    return Function(317, *args)

def SIN(*args):
    return Function(318, *args)

def SIND(*args):
    return Function(319, *args)

def SINH(*args):
    return Function(320, *args)

def SIZEOF(*args):
    return Function(321, *args)

def SLOPE_OF(*args):
    return Function(322, *args)

def SMOOTH(*args):
    return Function(323, *args)

def SOLVE(*args):
    return Function(324, *args)

def SORTVAL(*args):
    return Function(325, *args)

def SPACING(*args):
    return Function(326, *args)

def SPAWN(*args):
    return Function(327, *args)

def SPREAD(*args):
    return Function(328, *args)

def SQRT(*args):
    return Function(329, *args)

def SQUARE(*args):
    return Function(330, *args)

def STATEMENT(*args):
    return Function(331, *args)

def STD_DEV(*args):
    return Function(332, *args)

def STRING(*args):
    return Function(333, *args)

def STRING_OPCODE(*args):
    return Function(334, *args)

def SUBSCRIPT(*args):
    return Function(335, *args)

def SUBTRACT(*args):
    return Function(336, *args)

def SUM(*args):
    return Function(337, *args)

def SWITCH(*args):
    return Function(338, *args)

def SYSTEM_CLOCK(*args):
    return Function(339, *args)

def TAN(*args):
    return Function(340, *args)

def TAND(*args):
    return Function(341, *args)

def TANH(*args):
    return Function(342, *args)

def TASK_OF(*args):
    return Function(343, *args)

def TEXT(*args):
    return Function(344, *args)

def TIME_OUT_OF(*args):
    return Function(345, *args)

def TINY(*args):
    return Function(346, *args)

def TRANSFER(*args):
    return Function(347, *args)

def TRANSPOSE_(*args):
    return Function(348, *args)

def TRIM(*args):
    return Function(349, *args)

def UNARY_MINUS(*args):
    return Function(350, *args)

def UNARY_PLUS(*args):
    return Function(351, *args)

def UNION(*args):
    return Function(352, *args)

def UNITS(*args):
    return Function(353, *args)

def UNITS_OF(*args):
    return Function(354, *args)

def UNPACK(*args):
    return Function(355, *args)

def UNSIGNED(*args):
    return Function(356, *args)

def VAL(*args):
    return Function(357, *args)

def VALIDATION_OF(*args):
    return Function(358, *args)

def VALUE_OF(*args):
    return Function(359, *args)

def VAR(*args):
    return Function(360, *args)

def VECTOR(*args):
    return Function(361, *args)

def VERIFY(*args):
    return Function(362, *args)

def WAIT(*args):
    return Function(363, *args)

def WHEN_OF(*args):
    return Function(364, *args)

def WHERE(*args):
    return Function(365, *args)

def WHILE(*args):
    return Function(366, *args)

def WINDOW_OF(*args):
    return Function(367, *args)

def WORD(*args):
    return Function(368, *args)

def WORD_UNSIGNED(*args):
    return Function(369, *args)

def WRITE(*args):
    return Function(370, *args)

def ZERO(*args):
    return Function(371, *args)

def d2PI():
    return Function(372)

def dNARG():
    return Function(373)

def ELEMENT(*args):
    return Function(374, *args)

def RC_DROOP(*args):
    return Function(375, *args)

def RESET_PRIVATE():
    return Function(376)

def RESET_PUBLIC():
    return Function(377)

def SHOW_PRIVATE(*args):
    return Function(378, *args)

def SHOW_PUBLIC(*args):
    return Function(379, *args)

def SHOW_VM(*args):
    return Function(380, *args)

def TRANSLATE(*args):
    return Function(381, *args)

def TRANSPOSE_MUL(*args):
    return Function(382, *args)

def UPCASE(*args):
    return Function(383, *args)

def USING(*args):
    return Function(384, *args)

def VALIDATION(*args):
    return Function(385, *args)

def dDEFAULT():
    return Function(386)

def dEXPT():
    return Function(387)

def dSHOT():
    return Function(388)

def GETDBI(*args):
    return Function(389, *args)

def CULL(*args):
    return Function(390, *args)

def EXTEND(*args):
    return Function(391, *args)

def I_TO_X(*args):
    return Function(392, *args)

def X_TO_I(*args):
    return Function(393, *args)

def MAP(*args):
    return Function(394, *args)

def COMPILE_DEPENDENCY(*args):
    return Function(395, *args)

def DECOMPILE_DEPENDENCY(*args):
    return Function(396, *args)

def BUILD_CALL(*args):
    return Function(397, *args)

def ERRORLOGS_OF(*args):
    return Function(398, *args)

def PERFORMANCE_OF(*args):
    return Function(399, *args)

def XD(*args):
    return Function(400, *args)

def CONDITION_OF(*args):
    return Function(401, *args)

def SORT(*args):
    return Function(402, *args)

def dTHIS():
    return Function(403)

def DATA_WITH_UNITS(*args):
    return Function(404, *args)

def dATM():
    return Function(405)

def dEPSILON0():
    return Function(406)

def dGN():
    return Function(407)

def dMU0():
    return Function(408)

def EXTRACT(*args):
    return Function(409, *args)

def FINITE(*args):
    return Function(410, *args)

def BIT_SIZE(*args):
    return Function(411, *args)

def MODULO(*args):
    return Function(412, *args)

def SELECTED_INT_KIND(*args):
    return Function(413, *args)

def SELECTED_REAL_KIND(*args):
    return Function(414, *args)

def DSQL(*args):
    return Function(415, *args)

def ISQL(*args):
    return Function(416, *args)

def FTELL(*args):
    return Function(417, *args)

def MAKE_ACTION(*args):
    return Function(418, *args)

def MAKE_CONDITION(*args):
    return Function(419, *args)

def MAKE_CONGLOM(*args):
    return Function(420, *args)

def MAKE_DEPENDENCY(*args):
    return Function(421, *args)

def MAKE_DIM(*args):
    return Function(422, *args)

def MAKE_DISPATCH(*args):
    return Function(423, *args)

def MAKE_FUNCTION(*args):
    return Function(424, *args)

def MAKE_METHOD(*args):
    return Function(425, *args)

def MAKE_PARAM(*args):
    return Function(426, *args)

def MAKE_PROCEDURE(*args):
    return Function(427, *args)

def MAKE_PROGRAM(*args):
    return Function(428, *args)

def MAKE_RANGE(*args):
    return Function(429, *args)

def MAKE_ROUTINE(*args):
    return Function(430, *args)

def MAKE_SIGNAL(*args):
    return Function(431, *args)

def MAKE_WINDOW(*args):
    return Function(432, *args)

def MAKE_WITH_UNITS(*args):
    return Function(433, *args)

def MAKE_CALL(*args):
    return Function(434, *args)

def CLASS_OF(*args):
    return Function(435, *args)

def DSCPTR_OF(*args):
    return Function(436, *args)

def KIND_OF(*args):
    return Function(437, *args)

def NDESC_OF(*args):
    return Function(438, *args)

def ACCUMULATE(*args):
    return Function(439, *args)

def MAKE_SLOPE(*args):
    return Function(440, *args)

def REM(*args):
    return Function(441, *args)

def COMPLETION_MESSAGE_OF(*args):
    return Function(442, *args)

def INTERRUPT_OF(*args):
    return Function(443, *args)

def dSHOTNAME():
    return Function(444)

def BUILD_WITH_ERROR(*args):
    return Function(445, *args)

def ERROR_OF(*args):
    return Function(446, *args)

def MAKE_WITH_ERROR(*args):
    return Function(447, *args)

def DO_TASK(*args):
    return Function(448, *args)

def ISQL_SET(*args):
    return Function(449, *args)

def FS_FLOAT(*args):
    return Function(450, *args)

def FS_COMPLEX(*args):
    return Function(451, *args)

def FT_FLOAT(*args):
    return Function(452, *args)

def FT_COMPLEX(*args):
    return Function(453, *args)

def BUILD_OPAQUE(*args):
    return Function(454, *args)

def MAKE_OPAQUE(*args):
    return Function(455, *args)

def DICT(*args):
    return Function(456, *args)

def TUPLE(*args):
    return Function(457, *args)

def LIST(*args):
    return Function(458, *args)

def SQUEEZE(*args):
    return Function(459, *args)

def FLATTEN(*args):
    return Function(460, *args)

